# -*- coding: utf-8 -*-

import json
import enum

from .address import Address
from .aer import Aer
from .block_hash import BlockHash
from .event import Event
from .tx_hash import TxHash
from ..status.commit_status import CommitStatus
from ..status.tx_result_status import TxResultStatus
from ..grpc.rpc_pb2 import CommitResult
from ..grpc.blockchain_pb2 import Receipt
from ..utils.encoding import encode_tx_hash, encode_b58


@enum.unique
class TxResultType(enum.Enum):
    COMMIT_RESULT = 0
    RECEIPT = 1


class TxResult:
    def __init__(self, result, tx=None):
        self.tx = tx
        self.__result = result

        if type(result) == Receipt:
            self.__type = TxResultType.RECEIPT
            try:
                self.status = TxResultStatus(result.status)
                self.detail = result.ret
            except:
                self.status = TxResultStatus.ERROR
                self.detail = result.status
            self.contract_address = Address.encode(result.contractAddress)
            self.fee_used = Aer(result.feeUsed)
            self.cumulative_fee_used = Aer(result.cumulativeFeeUsed)
            self.bloom = result.bloom
            self.event_list = []
            for e in result.events:
                self.event_list.append(Event(e))
            self.block_no = result.blockNo
            self.block_hash = BlockHash(result.blockHash)
            self.tx_index = result.txIndex
            self.tx_hash = TxHash(result.txHash)
            self.from_address = Address.encode(getattr(result, 'from'))
            self.to_address = Address.encode(result.to)

            """
            if result.ret is None or 0 == len(result.ret):
                self.status = TxResultStatus.ERROR
                if 'CREATED' != result.status:
                    self.detail = result.status
            """
        elif type(result) == CommitResult:
            self.__type = TxResultType.COMMIT_RESULT
            self.tx_id = encode_tx_hash(self.__result.hash)
            self.status = CommitStatus(self.__result.error)
            self.detail = self.__result.detail
            self.contract_address = None
            self.fee_used = None
            self.cumulative_fee_used = None
            self.bloom = None
            self.event_list = []
            self.block_no = -1
            self.block_hash = None
            self.tx_index = -1
            self.tx_hash = None
            self.from_address = None
            self.to_address = None

    @property
    def type(self):
        return self.__type

    def json(self):
        return {
            'type': self.__type.name,
            'tx_id': self.tx_id,
            'status': self.status.name,
            'detail': self.detail,
            'contract_address': self.contract_address,
            'fee_used': str(self.fee_used) if self.fee_used is not None else None,
            'cumulative_fee_used': str(self.cumulative_fee_used) if self.cumulative_fee_used is not None else None,
            'bloom': encode_b58(self.bloom) if self.bloom is not None else None,
            'event_list': [event.json() for event in self.event_list],
            'block_no': self.block_no,
            'block_hash': str(self.block_hash) if self.block_hash is not None else None,
            'tx_index': self.tx_index,
            'tx_hash': str(self.tx_hash) if self.tx_hash is not None else None,
            'from_address': self.from_address,
            'to_address': self.to_address,
        }

    def __str__(self):
        return json.dumps(self.json(), indent=2)



# -*- coding: utf-8 -*-

"""Common utility module for converting types."""

import hashlib
import json
import toml
import socket
import ecdsa
import time

from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from ..obj.aergo_conf import AergoConfig
from ..grpc import blockchain_pb2
from ..constants import *
from .encoding import encode_b58


def convert_toml_to_aergo_conf(v):
    aergo_conf = AergoConfig()

    conf = toml.loads(v)
    for k, v in conf.items():
        if isinstance(v, dict):
            for k2, v2 in v.items():
                aergo_conf.add_conf(k2, v2, k)
        else:
            aergo_conf.add_conf(k, v)

    return aergo_conf


def convert_aergo_conf_to_toml(aergo_conf):
    return toml.dumps(aergo_conf.conf)


def convert_tx_to_grpc_tx(tx):
    grpc_tx = blockchain_pb2.Tx()
    grpc_tx.hash = bytes(tx.tx_hash)
    if tx.nonce is not None:
        grpc_tx.body.nonce = tx.nonce
    if tx.from_address is not None:
        grpc_tx.body.account = bytes(tx.from_address)
    if tx.to_address is not None:
        grpc_tx.body.recipient = bytes(tx.to_address)
    if tx.amount is not None:
        grpc_tx.body.amount = bytes(tx.amount)
    if tx.payload is not None:
        grpc_tx.body.payload = tx.payload
    grpc_tx.body.gasLimit = tx.fee_limit
    grpc_tx.body.gasPrice = bytes(tx.fee_price)
    grpc_tx.body.type = tx.tx_type.value
    grpc_tx.body.chainIdHash = tx.chain_id
    if tx.sign is not None:
        grpc_tx.body.sign = tx.sign
    return grpc_tx


def tx_to_grpc_tx(v):
    return convert_tx_to_grpc_tx(v)


def convert_tx_to_json(tx):
    if tx is None:
        return None

    return tx.json()


def tx_to_json(v):
    return convert_tx_to_json(v)


def convert_tx_to_formatted_json(tx):
    if tx is None:
        return None
    return json.dumps(convert_tx_to_json(tx), indent=2)


def tx_to_formatted_json(v):
    return convert_tx_to_formatted_json(v)


def convert_bytes_to_int_str(v):
    return ''.join('{:d} '.format(x) for x in v)


def bytes_to_int_str(v):
    return convert_bytes_to_int_str(v)


def convert_bytes_to_hex_str(v):
    return ''.join('0x{:02x} '.format(x) for x in v)


def convert_ip_bytes_to_str(v):
    if isinstance(v, str):
        return v

    l = len(v)

    # IPv4
    if 4 == l:
        return socket.inet_ntoa(v)
    elif 16 == l and all(v2 == 0 for v2 in list(v[:10])) and 255 == v[10] and 255 == v[11]:
        return socket.inet_ntoa(v[12:16])

    # IPv6
    return socket.inet_ntop(socket.AF_INET6, v)


def convert_bigint_to_bytes(number):
    q, r = divmod(len(bin(number))-2, 8)
    bytes_to_fit_number = q if r == 0 else q + 1
    return number.to_bytes(bytes_to_fit_number, 'big')


def bigint_to_bytes(v):
    return convert_bigint_to_bytes(v)


def convert_public_key_to_bytes(pubkey, curve=ecdsa.SECP256k1, compressed=True):
    if not isinstance(pubkey, ecdsa.ecdsa.Public_key):
        raise TypeError('value is not a valid public key')

    x = pubkey.point.x()
    x_bytes = ecdsa.util.number_to_string(x, curve.order)

    y = pubkey.point.y()
    if compressed:
        head = PUBLIC_KEY_COMPRESSED_E if 0 == y % 2 else PUBLIC_KEY_COMPRESSED_O
        y_bytes = b''
    else:
        head = PUBLIC_KEY_UNCOMPRESSED
        y_bytes = ecdsa.util.number_to_string(y, curve.order)

    return head + x_bytes + y_bytes


def public_key_to_bytes(pubkey, curve=ecdsa.SECP256k1, compressed=True):
    return convert_public_key_to_bytes(pubkey=pubkey,
                                       curve=curve,
                                       compressed=compressed)


def convert_bytes_to_public_key(v, curve=ecdsa.SECP256k1):
    if not isinstance(v, bytes):
        raise TypeError('value is not bytes')

    head = v[:1]
    if head not in (PUBLIC_KEY_UNCOMPRESSED,
                    PUBLIC_KEY_COMPRESSED_O,
                    PUBLIC_KEY_COMPRESSED_E):
        # can be a smart contract address, so no error
        raise ValueError("public key is not proper")

    x_bytes = v[1:curve.baselen+1]
    x = ecdsa.util.string_to_number(x_bytes)

    if PUBLIC_KEY_UNCOMPRESSED == head:
        y_bytes = v[curve.baselen+1:]
        y = ecdsa.util.string_to_number(y_bytes)
    else:
        a = curve.curve.a()
        b = curve.curve.b()
        p = curve.curve.p()

        if ecdsa.SECP256k1 == curve:
            # source: https://stackoverflow.com/questions/43629265/deriving-an-ecdsa-uncompressed-public-key-from-a-compressed-one?rq=1
            y_square = (pow(x, 3, p) + a * x + b) % p
            y_square_square_root = pow(y_square, (p+1)//4, p)
            if ((head == PUBLIC_KEY_COMPRESSED_E and y_square_square_root & 1) or
                    (head == PUBLIC_KEY_COMPRESSED_O and not y_square_square_root & 1)):
                y = (-y_square_square_root) % p
            else:
                y = y_square_square_root
        else:
            # if supporting more formula, need to implement
            assert 1 == 0

    point = ecdsa.ellipticcurve.Point(curve.curve, x, y, curve.order)
    return ecdsa.ecdsa.Public_key(curve.generator, point)


def bytes_to_public_key(v, curve=ecdsa.SECP256k1):
    return convert_bytes_to_public_key(v, curve=curve)


def encrypt_bytes(data, password):
    """
    https://cryptography.io/en/latest/hazmat/primitives/aead/
    :param data: bytes to encrypt
    :return: encrypted  data (bytes)
    """
    if isinstance(password, str):
        password = bytes(password, encoding='utf-8')

    m = hashlib.sha256()
    m.update(password)
    hash_pw = m.digest()

    m = hashlib.sha256()
    m.update(password)
    m.update(hash_pw)
    enc_key = m.digest()

    nonce = hash_pw[4:16]
    aesgcm = AESGCM(enc_key)
    return aesgcm.encrypt(nonce=nonce,
                          data=data,
                          associated_data=b'')

def decrypt_bytes(encrypted_bytes, password):
    """
    https://cryptography.io/en/latest/hazmat/primitives/aead/
    :param encrypted_bytes: encrypted data (bytes)
    :param password: to decrypt the exported bytes
    :return: decrypted bytes
    """
    if isinstance(password, str):
        password = password.encode('utf-8')

    m = hashlib.sha256()
    m.update(password)
    hash_pw = m.digest()

    m = hashlib.sha256()
    m.update(password)
    m.update(hash_pw)
    dec_key = m.digest()

    nonce = hash_pw[4:16]
    aesgcm = AESGCM(dec_key)
    dec_value = aesgcm.decrypt(nonce=nonce,
                               data=encrypted_bytes,
                               associated_data=b'')
    return dec_value


def get_hash(*strings, no_rand=False, no_encode=False):
    m = hashlib.sha256()
    if not no_rand:
        m.update(int(time.time()).to_bytes(8, 'little'))
    for string in strings:
        if isinstance(string, str):
            string = string.encode('utf-8')
        m.update(string)

    if no_encode:
        return m.digest()

    return encode_b58(m.digest())



import pytest

import ecdsa
import hashlib

from aergo.herapy.obj.address import Address
from aergo.herapy.obj.private_key import PrivateKey
from aergo.herapy.utils.signature import uncompress_key
from aergo.herapy.utils.encoding import encode_address, decode_address
from aergo.herapy.utils.converter import convert_public_key_to_bytes


def test_compare_pubkey():
    private_key = "6i2n6TAtsKBaWSujYFzqkBCiP834j1u3nwmffZb8dxCxDcQaXAb"
    address = "AmLnc5nhXjL3a3GUzv1Hb44LnGPvhZufDZ8s8oE9kazgbW6FgnUa"

    pubkey_compressed = decode_address(address).hex()
    pubkey_uncompressed = uncompress_key(pubkey_compressed)
    pubkey_str = bytes.fromhex(pubkey_uncompressed)
    pubkey_bytes = bytes.fromhex(pubkey_uncompressed)[1:]
    pubkey_bytes_x = pubkey_bytes[:ecdsa.SECP256k1.baselen]
    pubkey_bytes_y = pubkey_bytes[ecdsa.SECP256k1.baselen:]
    pubkey_x = ecdsa.util.string_to_number(pubkey_bytes_x)
    pubkey_y = ecdsa.util.string_to_number(pubkey_bytes_y)
    assert ecdsa.ecdsa.point_is_valid(ecdsa.SECP256k1.generator,
                                      pubkey_x, pubkey_y)

    pubkey_point = ecdsa.ellipticcurve.Point(ecdsa.SECP256k1.curve,
                                             pubkey_x, pubkey_y,
                                             ecdsa.SECP256k1.order)
    pubkey = ecdsa.ecdsa.Public_key(ecdsa.SECP256k1.generator, pubkey_point)

    pk = PrivateKey(pk=private_key)
    pk_pubkey = pk.public_key
    pk_pubkey_str = convert_public_key_to_bytes(pk.public_key,
                                                compressed=False)
    pk_pubkey_c_str = convert_public_key_to_bytes(pk.public_key,
                                                  compressed=True)

    assert address == str(pk.address)
    assert pubkey.point == pk_pubkey.point
    assert pubkey_str == pk_pubkey_str
    assert address == encode_address(pk_pubkey_c_str)

    addr = Address(None, empty=True)
    addr.value = address
    addr_pubkey = addr.public_key
    addr_pubkey_str = convert_public_key_to_bytes(addr.public_key,
                                                  compressed=False)
    addr_pubkey_c_str = convert_public_key_to_bytes(addr.public_key,
                                                    compressed=True)

    assert pubkey.point == addr_pubkey.point
    assert pubkey_str == addr_pubkey_str
    assert address == encode_address(addr_pubkey_c_str)



# -*- coding: utf-8 -*-

import hashlib
import json

from google.protobuf.json_format import MessageToJson
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from .obj import address as addr
from .obj import aer
from .obj import private_key as pk

from .utils.encoding import decode_address, decode_root
from .utils import merkle_proof as mp
from .utils.converter import encrypt_bytes, decrypt_bytes

from .errors.general_exception import GeneralException


class Account:
    """ Account can be a user account with private and public key,
    or a contract account.
    """
    def __init__(self, private_key=None, empty=False):
        if empty:
            self.__private_key = None
            self.__address = None
            self.__state = None
            self.__state_proof = None
            return

        self.__private_key = pk.PrivateKey(private_key)
        self.__address = self.__private_key.address
        self.__state = None
        self.__state_proof = None

    def __str__(self):
        state = self.state
        if self.__state_proof:
            state = MessageToJson(self.__state_proof)

        if state:
            state = json.loads(state)

        account = {
            "address": str(self.__address),
            "balance": str(self.balance),
            "nonce": str(self.nonce),
            "state": state,
        }
        return json.dumps(account, indent=2)

    def sign_msg_hash(self, msg_hash):
        return self.__private_key.sign_msg(msg_hash)

    def verify_sign(self, msg_hash, sign):
        return self.__private_key.verify_sign(msg_hash, sign)

    @property
    def private_key(self):
        return self.__private_key

    @property
    def public_key(self):
        if self.__private_key is not None:
            return self.__private_key.public_key

        if self.__address is not None:
            return self.__address.public_key

        return None

    @property
    def address(self):
        return self.__address

    @address.setter
    def address(self, v):
        if self.__address is not None:
            raise ValueError('not empty account')

        self.__address = addr.Address(None, empty=True)
        self.__address.value = v

    @property
    def state(self):
        if self.__state is None:
            return None
        return MessageToJson(self.__state)

    @state.setter
    def state(self, v):
        self.__state = v

    @property
    def state_proof(self):
        return self.__state_proof

    @state_proof.setter
    def state_proof(self, v):
        self.__state_proof = v

    @property
    def nonce(self):
        if self.__state is None:
            return -1
        return self.__state.nonce

    @nonce.setter
    def nonce(self, v):
        if self.__state.nonce > v:
            raise ValueError("the nonce value should be bigger than the current nonce value: {}".format(self.__state.nonce))
        self.__state.nonce = v

    @property
    def balance(self):
        if self.__state is None:
            return aer.Aer()
        return aer.Aer(int.from_bytes(self.__state.balance, 'big'))

    @property
    def code_hash(self):
        if self.__state is None:
            return None
        return self.__state.codeHash

    @property
    def storage_root(self):
        if self.__state is None:
            return None
        return self.__state.storageRoot

    @property
    def sql_recovery_point(self):
        if self.__state is None:
            return None
        return self.__state.sqlRecoveryPoint

    @staticmethod
    def encrypt_account(account, password):
        """
        https://cryptography.io/en/latest/hazmat/primitives/aead/
        :param account: account to export
        :return: encrypted account data (bytes)
        """
        return encrypt_bytes(bytes(account.private_key), password)

    @staticmethod
    def decrypt_account(encrypted_bytes, password):
        """
        https://cryptography.io/en/latest/hazmat/primitives/aead/
        :param encrypted_bytes: encrypted data (bytes) of account
        :param password: to decrypt the exported bytes
        :return: account instance
        """
        try:
            dec_value = decrypt_bytes(encrypted_bytes, password)
        except InvalidTag as e:
            raise GeneralException("Fail to decrypt an account. Please check the password.") from e
        return Account(private_key=dec_value)

    def verify_proof(self, root):
        """ verify that the given inclusion and exclusion proofs are correct """
        if self.__state_proof is None:
            return False

        if isinstance(root, str) and len(root) != 0:
            root = decode_root(root)

        if bytes(self.__address) != self.__state_proof.key:
            return False

        trie_key = hashlib.sha256(bytes(self.address)).digest()
        value = hashlib.sha256(self.__state_proof.state.SerializeToString()).digest()
        ap = self.__state_proof.auditPath

        if self.__state_proof.bitmap:
            height = self.__state_proof.height
            bitmap = self.__state_proof.bitmap
            if self.__state_proof.inclusion:
                return mp.verify_inclusion_c(ap, height, bitmap, root, trie_key, value)
            else:
                return mp.verify_exclusion_c(root, ap, height, bitmap, trie_key,
                                             self.__state_proof.proofKey,
                                             self.__state_proof.proofVal)
        else:
            if self.__state_proof.inclusion:
                return mp.verify_inclusion(ap, root, trie_key, value)
            else:
                return mp.verify_exclusion(root, ap, trie_key,
                                           self.__state_proof.proofKey,
                                           self.__state_proof.proofVal)



# -*- coding: utf-8 -*-

import json

from .block_hash import BlockHash
from .consensus_info import ConsensusInfo
from ..utils.encoding import encode_b58


class BlockchainStatus:
    def __init__(self, status):
        self._status = status
        self._bbh = BlockHash(self._status.best_block_hash)
        self._consensus_info = ConsensusInfo(self._status.consensus_info)

    @property
    def best_block_hash(self):
        return self._bbh

    @property
    def best_block_height(self):
        return self._status.best_height

    @property
    def best_chain_id_hash(self):
        return self._status.best_chain_id_hash

    @property
    def best_chain_id_hash_b58(self):
        return encode_b58(self._status.best_chain_id_hash)

    @property
    def consensus_info(self):
        return self._consensus_info

    def json(self):
        return {
            "best_block_hash": str(self.best_block_hash),
            "best_block_height": self.best_block_height,
            "best_chain_id_hash": self.best_chain_id_hash_b58,
            "consensus_info": self.consensus_info.json(),
        }

    def __str__(self):
        return json.dumps(self.json(), indent=2)


import pytest

import decimal


from aergo.herapy.obj.block import Block
from aergo.herapy.obj.block_hash import BlockHash
from aergo.herapy.obj.transaction import TxType
from aergo.herapy.utils.encoding import encode_block_hash, decode_block_hash,\
    encode_tx_hash, decode_tx_hash, encode_payload, decode_payload,\
    encode_b58, decode_b58, encode_address
from aergo.herapy.utils.converter import bigint_to_bytes, get_hash
from aergo.herapy.grpc import blockchain_pb2


def test_fail():
    with pytest.raises(TypeError):
        bh = BlockHash(1234)
        str(bh)


def test_blockhash():
    bh = BlockHash('test')
    assert bytes(bh) == decode_block_hash('test')
    assert bytes(bh) == bh.value
    assert str(bh) == 'test'

    bh = BlockHash(b'test')
    assert bytes(bh) == b'test'
    assert bytes(bh) == bh.value
    assert str(bh) == encode_block_hash(b'test')


def test_grpc_block():
    grpc_block = blockchain_pb2.Block()
    grpc_block.hash = b'block_hash'
    # header
    grpc_block.header.chainID = b'chain_id'
    grpc_block.header.prevBlockHash = b'prev_block_hash'
    grpc_block.header.blockNo = 123
    grpc_block.header.timestamp = 12345
    grpc_block.header.blocksRootHash = b'blocks_root_hash'
    grpc_block.header.txsRootHash = b'txs_root_hash'
    grpc_block.header.receiptsRootHash = b'receipts_root_hash'
    grpc_block.header.confirms = 10
    grpc_block.header.pubKey = b'pub_key'
    grpc_block.header.coinbaseAccount = b'coinbase_account'
    grpc_block.header.sign = b'block_sign'
    # body
    tx1 = blockchain_pb2.Tx()
    tx1.hash = b'tx1_hash'
    tx1.body.nonce = 10
    tx1.body.account = b'yp'
    tx1.body.recipient = b'daniel'
    tx1.body.amount = bigint_to_bytes(1)
    tx1.body.payload = b'tx1_payload'
    tx1.body.gasLimit = 1
    tx1.body.gasPrice = bigint_to_bytes(1)
    tx1.body.type = blockchain_pb2.GOVERNANCE
    tx1.body.sign = b'tx1_sign'
    tx2 = blockchain_pb2.Tx()
    tx2.hash = b'tx2_hash'
    tx2.body.nonce = 356
    tx2.body.account = b'daniel'
    tx2.body.recipient = b'hannah'
    tx2.body.amount = bigint_to_bytes(2)
    #tx2.body.payload = b'tx2_payload'
    tx2.body.gasLimit = 2
    tx2.body.gasPrice = bigint_to_bytes(2)
    tx2.body.type = blockchain_pb2.NORMAL
    tx2.body.sign = b'tx2_sign'
    grpc_block.body.txs.extend([tx1, tx2,])

    block = Block(grpc_block=grpc_block)
    assert block.hash.value == b'block_hash'
    assert block.chain_id == b'chain_id'
    assert block.height == 123
    assert block.block_no == block.height
    assert bytes(block.prev.hash) == b'prev_block_hash'
    assert block.timestamp == 12345
    assert block.blocks_root_hash == b'blocks_root_hash'
    assert block.txs_root_hash == b'txs_root_hash'
    assert block.receipts_root_hash == b'receipts_root_hash'
    assert block.confirms == 10
    assert block.public_key == b'pub_key'
    assert block.coinbase_account == b'coinbase_account'
    assert block.sign == b'block_sign'
    assert len(block.tx_list) == 2
    assert block.num_of_tx == 2
    assert block.size == 254
    block_json = block.json(header_only=True)
    assert 'body' not in block_json
    block_json = block.json(header_only=False)
    assert 'body' in block_json
    assert 'tx_list' in block_json['body']
    assert 'hash' in block_json
    assert block_json['hash'] == encode_block_hash(b'block_hash')
    assert 'header' in block_json
    assert block_json['header']['chain_id'] == get_hash(b'chain_id', no_rand=True, no_encode=False)
    assert block_json['header']['previous_block_hash'] == encode_block_hash(b'prev_block_hash')
    assert block_json['header']['blocks_root_hash'] == encode_b58(b'blocks_root_hash')
    assert block_json['header']['txs_root_hash'] == encode_b58(b'txs_root_hash')
    assert block_json['header']['receipts_root_hash'] == encode_b58(b'receipts_root_hash')
    assert block_json['header']['block_no'] == 123
    assert block_json['header']['timestamp'] == 12345
    assert block_json['header']['confirms'] == 10
    assert block_json['header']['pub_key'] == encode_b58(b'pub_key')
    assert block_json['header']['sign'] == encode_b58(b'block_sign')
    assert block_json['header']['coinbase_account'] == encode_address(b'coinbase_account')
    block_str = str(block)
    assert 'hash' in block_str
    assert 'header' in block_str
    assert 'body' in block_str

    # check tx1
    block_tx1 = block.get_tx(0)
    assert block == block_tx1.block
    assert block_tx1.index_in_block == 0
    assert block_tx1.is_in_mempool == False
    assert block_tx1.nonce == 10
    block_tx1.nonce = 123
    assert block_tx1.nonce != 123
    assert block_tx1.nonce == 10
    assert bytes(block_tx1.tx_hash) == b'tx1_hash'
    assert str(block_tx1.tx_hash) == encode_tx_hash(b'tx1_hash')
    assert bytes(block_tx1.from_address) == b'yp'
    block_tx1.from_address = b'hannah'
    assert bytes(block_tx1.from_address) != b'hannah'
    assert bytes(block_tx1.from_address) == b'yp'
    assert bytes(block_tx1.to_address) == b'daniel'
    block_tx1.to_address = b'hannah'
    assert bytes(block_tx1.to_address) != b'hannah'
    assert bytes(block_tx1.to_address) == b'daniel'
    assert block_tx1.amount.aer == '1 aer'
    assert block_tx1.amount.gaer == '0.000000001 gaer'
    assert block_tx1.amount.aergo == '0.000000000000000001 aergo'
    block_tx1.amount = '100 aer'
    assert block_tx1.amount.aer != '100 aer'
    assert block_tx1.amount.aer == '1 aer'
    assert block_tx1.payload == b'tx1_payload'
    block_tx1.payload = b'new_payload'
    assert block_tx1.payload != b'new_payload'
    assert block_tx1.payload == b'tx1_payload'
    assert block_tx1.payload_str == encode_payload(b'tx1_payload')
    assert block_tx1.tx_type == TxType.GOVERNANCE
    block_tx1.tx_type = TxType.NORMAL
    assert block_tx1.tx_type != TxType.NORMAL
    assert block_tx1.tx_type == TxType.GOVERNANCE
    # check tx2
    block_tx2 = block.get_tx(1)
    assert block == block_tx2.block
    assert block_tx2.index_in_block == 1
    assert block_tx2.is_in_mempool == False
    assert block_tx2.nonce ==356
    block_tx2.nonce = 123
    assert block_tx2.nonce != 123
    assert block_tx2.nonce == 356
    assert bytes(block_tx2.tx_hash) == b'tx2_hash'
    assert str(block_tx2.tx_hash) == encode_tx_hash(b'tx2_hash')
    assert bytes(block_tx2.from_address) == b'daniel'
    block_tx2.from_address = b'yp'
    assert bytes(block_tx2.from_address) != b'yp'
    assert bytes(block_tx2.from_address) == b'daniel'
    assert bytes(block_tx2.to_address) == b'hannah'
    block_tx2.to_address = b'yp'
    assert bytes(block_tx2.to_address) != b'yp'
    assert bytes(block_tx2.to_address) == b'hannah'
    assert block_tx2.amount.aer == '2 aer'
    assert block_tx2.amount.gaer == '0.000000002 gaer'
    assert block_tx2.amount.aergo == '0.000000000000000002 aergo'
    block_tx2.amount = '100 aer'
    assert block_tx2.amount.aer != '100 aer'
    assert block_tx2.amount.aer == '2 aer'
    assert block_tx2.payload == b''
    block_tx2.payload = b'new_payload'
    assert block_tx2.payload != b'new_payload'
    assert block_tx2.payload == b''
    assert block_tx2.payload_str == None
    assert block_tx2.tx_type == TxType.NORMAL
    block_tx2.tx_type = TxType.GOVERNANCE
    assert block_tx2.tx_type != TxType.GOVERNANCE
    assert block_tx2.tx_type == TxType.NORMAL

def test_block():
    bh = BlockHash('test')



import sys
import traceback
import time

from threading import Thread

import aergo.herapy as herapy


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    print(*args, **kwargs)


def read_movie_event_stream(movie_event_stream):
    while True:
        try:
            print("RECEIVE: 'movie' Event: {}".format(str(next(movie_event_stream))))
            print(movie_event_stream.is_active())
            print(movie_event_stream.cancelled())
            print(movie_event_stream.done())
            print(movie_event_stream.running())
        except Exception as e:
            print(e)
            print(movie_event_stream.is_active())
            print(movie_event_stream.cancelled())
            print(movie_event_stream.done())
            print(movie_event_stream.running())
            break

    print("END 'movie' event stream")
    movie_event_stream.stop()


def run():
    print("------ Payload -----------")
    """

function constructor(name)
        contract.event("movie", "robocop", 1, "I'll be back")
end

function movie_event(name)
        contract.event("movie", "arnold", 2, "It's mine")
        contract.event("movie", "terminator", 3, "No, it's mine")
end

function ani_event(name)
        contract.event("a-time!!")
        contract.event("a-time!!", "pinn", "Adventure time!!", false, bignum.number("999999999999999"), {"princess bubble gum", 1})
        contract.event("a-time!!", "Jake", "Adventure time!!", false, {"whatever", 1, {true, false}, bignum.number("999999999999999")})

        contract.event("movie", "yoda", 3, "i'm not the ani character.")
end

abi.register(movie_event, ani_event)
    """
    payload_str = "D1ZsfUyj4q8GA6AtbW81TzSTSvqaeUvcsePsaZufptRK5qSb83fyXRWdEgGhLWs17JfRNrM4iLjGe54fkyvsyjHVgFCHqTzJWK4Yz8uQwRqzF5x1g1twZ3jPP1XtJ95y43RD48qXGoeeoDsmmNErQJCux6d4eLZwE344bz5hM1DdWEd22yE6ogC37RLRtRZ5FFKm3HkxaM67G7VqWfsX5t2Tmd23Myn62hkAdNWjSBaWizqyi6ENB3VD2YJGG82bVeq9a4zMQdK58Wxnfcfqo14Vo9Y9jsY8A5QYNRi6uotRfiYt3LFrPjPBWgyQRn4qFQypSW5vfgFZvv9VBdFPGgoMN5zeBp7VNWXvE5o8CGAf3fuFu2gCGLS4DfCtQe7VNnsCUi7U1AdmdupqoxMHM6kCfEtk77YyiaaSTm4xUSt7R2hwzs2Xnfh8CLxMaH2cg3rsWF4fefs5KvnjN3F1q43og6BueK66o8NssyY41kAtincbYhDC8oHaBkMTdXD6mkhb3QECwn8dh8k9a3KTys1tdUzf99nqpK5YrAcZFDjbrmmmtcBYwZg6zdRAkQnJptvr25HzeHMfbfERH8P1XzYmjqYynFvUNv4utuWZ7avS2xvJoZph3nSa8Xuhbdw6TFLVCDZVfZB6j8kQbJnb1DoEEtvhPHLMP4StcgiE9grGQuduZ5iHYtS8Nk696Fq6Y9gzp19bWD3VKarq1XuueM6FNrzCQNY22nov4mxbM8vcxp4axdjmZNWLcdttFycazorxCNbn1YBQFzGPGzstJPN72RB9dxV17cb1TuREKTF7mVLLeahgHsY4eVnA1m5Z8U7tFKswRDTGeXv6vCt7MstVBztdd4sRfHTXKdFCgfoQk2zwzGS8FEVgxAP3Y4Tip4nuQd8hVyTiaApFmNnkD9byxUNCc44Tr9LAnRycc239HGpZjWBwsJvmb4jsY2fuEELXfSgQf4yZhnbcPC4DYPQLDRXc8E6DdLzixMk7cAfUZqododk9SWzjB9FSKtzVYWiKuyg2uZoRaxDmrtZpe6pXL8F4pzt2P2HEG6jyFn5MT9PBWhkj11vqAjWpdRgNAbTKEb5zk9ZeER9YUAT8etk2VFpAGqmGnv9CTEQr2rfDQeNgNYzoy44mNZonhi4n4zo9h2HmrSsatWa8oxbJJYihGs7zD65TiqRfnKg43Xx5qD"
    payload = herapy.utils.decode_address(payload_str)
    print(''.join('{:d} '.format(x) for x in payload))

    byte_code_len = int.from_bytes(payload[:4], byteorder='little')
    print("payload: byte code length = ", byte_code_len)

    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Set Sender Account -----------")
        sender_private_key = "6hbRWgddqcg2ZHE5NipM1xgwBDAKqLnCKhGvADWrWE18xAbX8sW"
        sender_account = aergo.new_account(private_key=sender_private_key)
        print("  > Sender Address: {}".format(sender_account.address))
        print(herapy.utils.convert_bytes_to_int_str(bytes(sender_account.address)))

        print("------ Deploy SC -----------")
        tx, result = aergo.deploy_sc(amount=0, payload=payload, args=1234)
        print("  > TX: {}".format(tx.tx_hash))
        print("{}".format(herapy.utils.convert_tx_to_json(tx)))
        if result.status != herapy.CommitStatus.TX_OK:
            eprint("    > ERROR[{0}]: {1}".format(result.status, result.detail))
            aergo.disconnect()
            return
        else:
            print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))
            print(herapy.utils.convert_bytes_to_int_str(bytes(tx.tx_hash)))

        time.sleep(3)

        print("------ Check deployment of SC -----------")
        print("  > TX: {}".format(tx.tx_hash))
        result = aergo.get_tx_result(tx.tx_hash)
        if result.status != herapy.TxResultStatus.CREATED:
            eprint("  > ERROR[{0}]:{1}: {2}".format(
                result.contract_address, result.status, result.detail))
            aergo.disconnect()
            return

        sc_address = result.contract_address
        print("  > SC Address: {}".format(sc_address))

        print("------ Get reading events threads -----------")
        movie_event_stream = aergo.receive_event_stream(sc_address, "movie",
                                                        start_block_no=0)
        movie_thread = Thread(target=read_movie_event_stream,
                              args=[movie_event_stream])
        movie_thread.start()

        print("------ Call SC -----------")
        tx, result = aergo.call_sc(sc_address, "movie_event")
        print("movie_event tx hash: {}".format(str(tx.tx_hash)))
        time.sleep(3)

        print("------ Cancel 'movie' event stream -----------")
        movie_event_stream.cancel()

        while True:
            if movie_event_stream.stopped:
                break
            time.sleep(3)

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception as e:
        eprint(e)
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-

from .stream import Stream
from .event import Event


class EventStream(Stream):
    def __init__(self, event_stream):
        super().__init__(event_stream)

    def next(self):
        return Event(grpc_event=next(self._grpc_stream))



# -*- coding: utf-8 -*-

import json

from .aer import Aer
from .chain_id import ChainID
from .consensus_info import ConsensusInfo


class NodeInfo:
    def __init__(self, node_info):
        self._node_info = node_info

        self._status_map = {}
        for k, v in node_info.status.items():
            self._status_map[k] = v
            setattr(self, "status_{}".format(k), v)

        self._config_map = {}
        for k, v in node_info.config.items():
            if k not in self._config_map:
                self._config_map[k] = {}
            for k2, v2 in v.props.items():
                self._config_map[k][k2] = v2
                setattr(self, "config_{}_{}".format(k, k2), v2)

    def json(self):
        ret = {
            "status": {},
            "config": {},
        }

        for k, v in self._status_map.items():
            ret['status'][k] = v
        for k, v in self._config_map.items():
            ret['config'][k] = v

        return ret

    def __str__(self):
        return json.dumps(self.json(), indent=2)


# -*- coding: utf-8 -*-

import hashlib

from google.protobuf.json_format import MessageToJson
from aergo.herapy.utils import merkle_proof as mp


class VarProofs(list):
    """ VarProof holds the inclusion/exclusion proof of a variable state
    inside a contract state trie
    """
    def __init__(self, var_proofs, storage_keys):
        self.__var_proofs = var_proofs
        self.__storage_keys = storage_keys
        list.__init__(self, var_proofs[:])

    def __str__(self):
        json_string = ""
        for proof in self.__var_proofs:
            json_string += MessageToJson(proof)
        return json_string

    @property
    def var_proofs(self):
        return self.__var_proofs

    @property
    def storage_keys(self):
        return self.__storage_keys

    def verify_var_proof(self, root, var_proof, trie_key):
        if var_proof.key != trie_key:
            return False
        value = hashlib.sha256(var_proof.value).digest()
        ap = var_proof.auditPath

        if var_proof.bitmap:
            height = var_proof.height
            bitmap = var_proof.bitmap

            if var_proof.inclusion:
                return mp.verify_inclusion_c(ap, height, bitmap, root, trie_key, value)
            else:
                return mp.verify_exclusion_c(root, ap, height, bitmap, trie_key,
                                             var_proof.proofKey,
                                             var_proof.proofVal)
        else:
            if var_proof.inclusion:
                return mp.verify_inclusion(ap, root, trie_key, value)
            else:
                return mp.verify_exclusion(root, ap, trie_key,
                                           var_proof.proofKey,
                                           var_proof.proofVal)

    def verify_proof(self, root):
        """verify that the given inclusion and exclusion proofs are correct """
        if self.__var_proofs is None or 0 == self.__len__():
            return False

        for i, storage_proof in enumerate(self.__var_proofs):
            if not self.verify_var_proof(root, storage_proof, self.storage_keys[i]):
                return False
        return True



import sys
import traceback

import aergo.herapy as herapy


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    print(*args, **kwargs)


def run():
    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845',
                      tls_ca_cert="./cert/rootca.crt",
                      tls_cert="./cert/client.pem",
                      tls_key="./cert/client.key")

        print("------ Get Node Info -----------")
        node_state = aergo.get_node_state()
        print("Node State: {}".format(str(node_state)))
        node_info = aergo.get_node_info()
        print("Node Info: {}".format(str(node_info)))

        print("------ Get Blockchain Info -----------")
        blockchain_info = aergo.get_chain_info()
        print("Blockchain Info: {}".format(str(blockchain_info)))
        blockchain_info = aergo.get_chain_info(with_consensus_info=False)
        print("Blockchain Info: {}".format(str(blockchain_info)))

        print("------ Get Consensus Info -----------")
        consensus_info = aergo.get_consensus_info()
        print("Consensus Info: {}".format(str(consensus_info)))

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception as e:
        eprint(e)
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()
